<?php

/**
 * @file
 * PubsubHubbub library.
 */

// TTL of a subscription - 6 months.
define('PUSH_HUB_LEASE_SECONDS', 15768000);

/**
 * A PubsubHubbub Hub.
 */
class PuSHHub {
  // Class to use for storing subscriptions.
  protected $subscriptions;

  /**
   * Singleton.
   */
  public function instance(PuSHHubSubscriptionsInterface $subscriptions) {
    static $hub;
    if (empty($hub)) {
      $hub = new PuSHHub($subscriptions);
    }
    return $hub;
  }

  /**
   * Constructor.
   *
   * @param $subscriptions
   *   Subsriptions object that handles subscription storage. Must implement
   *   PuSHHubSubscriptionsInterface.
   */
  protected function __construct(PuSHHubSubscriptionsInterface $subscriptions) {
    $this->subscriptions = $subscriptions;
  }

  /**
   * Get all subscribers for a given topic.
   */
  public function allSubscribers($topic) {
    return $this->subscriptions->all($topic);
  }

  /**
   * Notify a specific subscriber of a change in a topic. API user is
   * responsible for handling failed requests.
   *
   * @param $topic
   *   The topic that changed.
   * @param $subscriber
   *   A subscriber callback.
   * @param $changed
   *   The full or partial feed that contains the changed elements. If NULL,
   *   a light ping (without any content) will be issued to subscriber and
   *   subscriber is expected to fetch content from publisher. Light pings are
   *   not pubsubhubbub spec conform.
   *
   * @return
   *   TRUE if subscriber was successfully notified, FALSE otherwise.
   */
  public function notify($topic, $subscriber, $changed = NULL) {
    if ($changed === NULL) {
      return $this->lightPing($subscriber);
    }
    return $this->fatPing($subscriber, $changed, $this->subscriptions->secret($topic, $subscriber));
  }

  /**
   * Verify subscription request.
   */
  public function verify($post) {
    // Send verification request to subscriber.
    $challenge = md5(session_id() . rand());
    $query = array(
      'hub.mode=' . $post['hub_mode'],
      'hub.topic=' . urlencode($post['hub_topic']),
      'hub.challenge=' . $challenge,
      'hub.lease_seconds=' . PUSH_HUB_LEASE_SECONDS, // disregard what's been requested ($post['hub_lease_seconds'])
      'hub.verify_token=' . $post['hub_verify_token'],
    );
    $parsed = parse_url($post['hub_callback']);
    $request = curl_init($post['hub_callback'] . (empty($parsed['query']) ? '?' : '&') . implode('&', $query));
    curl_setopt($request, CURLOPT_FOLLOWLOCATION, TRUE);
    curl_setopt($request, CURLOPT_RETURNTRANSFER, TRUE);
    $data = curl_exec($request);
    $code = curl_getinfo($request, CURLINFO_HTTP_CODE);
    curl_close($request);
    if ($code > 199 && $code < 300 && $data == $challenge) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Handle a subscription request.
   *
   * @param array $post
   *   A valid PubSubHubbub subscribe request.
   *
   *   The keys of this array are defined at http://pubsubhubbub.googlecode.com/svn/trunk/pubsubhubbub-core-0.3.html#anchor5
   *
   *   - 'hub_callback': REQUIRED. The subscriber's callback URL where
   *     notifications should be delivered.
   *   - 'hub_mode': REQUIRED. The literal string "subscribe".
   *   - 'hub_topic': REQUIRED. The topic URL that the subscriber wishes to
   *     subscribe to.
   *   - 'hub_secret': OPTIONAL. A subscriber-provided secret string that will
   *     be used to compute an HMAC digest for authorized content distribution.
   *     If not supplied, the HMAC digest will not be present for content
   *     distribution requests. This parameter SHOULD only be specified when the
   *     request was made over HTTPS [RFC2818]. This parameter MUST be less than
   *     200 bytes in length.
   *   - 'hub_verify_token': OPTIONAL. A subscriber-provided opaque token that
   *     will be echoed back in the verification request to assist the subscriber
   *     in identifying which subscription request is being verified. If this is
   *     not included, no token will be included in the verification request.
   *
   *   This Push Hub Drupal module deviates from the spec in it's usage of hub_verify.
   *   The spec intends this parameter use values of 'sync' and/or 'async' to
   *   determine how the hub will verify the subscription. This module verifies
   *   immediately.
   *
   *   The module maintains spec conformance while ignoring incoming values for
   *   hub.lease_seconds. The constant PUSH_HUB_LEASE_SECONDS is used instead.
   *
   * @return
   *   Boolean TRUE or FALSE for whether the subscription was successful or unsuccessful.
   */
  public function subscribe(array $post) {
    watchdog('push_hub', "subscribe request", NULL, WATCHDOG_DEBUG);
    if ($post['hub_mode'] === 'subscribe' && isset($post['hub_topic']) && isset($post['hub_callback'])  && $this->verify($post)) {
      $this->subscriptions->save($post['hub_topic'], $post['hub_callback'], isset($post['hub_secret']) ? $post['hub_secret'] : '');
      watchdog('push_hub', "subscription verified and saved", NULL, WATCHDOG_DEBUG);
      return TRUE;
    } else {
      watchdog('push_hub', "subscription invalid and not saved", NULL, WATCHDOG_DEBUG);
      return FALSE;
    }
  }

  /**
   * Handle a subscription request.
   *
   *   The keys of this array are defined at http://pubsubhubbub.googlecode.com/svn/trunk/pubsubhubbub-core-0.3.html#anchor5
   *
   *   - 'hub_callback': REQUIRED. The subscriber's callback URL where
   *     notifications should be delivered.
   *   - 'hub_mode': REQUIRED. The literal string "unsubscribe".
   *   - 'hub_topic': REQUIRED. The topic URL that the subscriber wishes to
   *     subscribe to.
   *   - 'hub_secret': OPTIONAL. A subscriber-provided secret string that will
   *     be used to compute an HMAC digest for authorized content distribution.
   *     If not supplied, the HMAC digest will not be present for content
   *     distribution requests. This parameter SHOULD only be specified when the
   *     request was made over HTTPS [RFC2818]. This parameter MUST be less than
   *     200 bytes in length.
   *   - 'hub_verify_token': OPTIONAL. A subscriber-provided opaque token that
   *     will be echoed back in the verification request to assist the subscriber
   *     in identifying which subscription request is being verified. If this is
   *     not included, no token will be included in the verification request.
   *
   *   This Push Hub Drupal module deviates from the spec in it's usage of hub_verify.
   *   The spec intends this parameter use values of 'sync' and/or 'async' to
   *   determine how the hub will verify the subscription. This module verifies
   *   immediately.
   *
   *   The module maintains spec conformance while ignoring incoming values for
   *   hub.lease_seconds. The constant PUSH_HUB_LEASE_SECONDS is used instead.
   *
   * @return
   *   Boolean TRUE or FALSE for whether the subscription was successful or unsuccessful.
   */
  public function unsubscribe(array $post) {
    watchdog('push_hub', "unsubscribe request", NULL, WATCHDOG_DEBUG);
    if ($post['hub_mode'] === 'unsubscribe' && isset($post['hub_topic']) && isset($post['hub_callback'])  && $this->verify($post)) {
      // @todo, does delete need to return anything to verify that the delete actually happened?
      $this->subscriptions->delete($post['hub_topic'], $post['hub_callback']);
      watchdog('push_hub', "unsubscription verified and record deleted", NULL, WATCHDOG_DEBUG);
      return TRUE;
    } else {
      watchdog('push_hub', "unsubscription invalid and no records deleted", NULL, WATCHDOG_DEBUG);
      return FALSE;
    }
  }

  /**
   * Expire old subscriptions.
   */
  public function expire() {
    $this->subscriptions->expire(PUSH_HUB_LEASE_SECONDS);
  }

  /**
   * Helper for posting a fat ping. Uses drupal_http_request().
   */
  protected function fatPing($url, $content, $secret) {
    $result = FALSE;

    // Build up an array of options for drupal_http_request();
    $options = array(
      'method' => 'POST',
      'headers' => array(
        'Content-Type' => 'application/atom+xml',
         'X-Hub-Signature' => "sha1=" . hash_hmac('sha1', $content, $secret),
      ),
      'data' => $content,
    );

    $response = drupal_http_request($url, $options);

    if (!empty($response->code) && $response->code >= 200 && $response->code < 300) {
      $result = TRUE;
    }

    return $result;
  }

  /**
   * Issue a light ping, not spec conform.
   */
  protected function lightPing($url) {
    $request = curl_init($url);
    curl_setopt($request, CURLOPT_FOLLOWLOCATION, TRUE);
    curl_setopt($request, CURLOPT_RETURNTRANSFER, TRUE);
    $data = curl_exec($request);
    $code = curl_getinfo($request, CURLINFO_HTTP_CODE);
    curl_close($request);
    if ($code >= 200 && $code < 300) {
      return TRUE;
    }
    return FALSE;
  }
}

/**
 * Describes the storage for PuSHHub subscriptions. Implement to provide a
 * storage class.
 */
interface PuSHHubSubscriptionsInterface {
  /**
   * Get the secret for a given topic/subscriber pair.
   *
   * @return
   *   Shared secret to sign a notification with.
   */
  public function secret($topic, $subscriber);

  /**
   * Save a subscription.
   *
   * @param $topic
   *   A topic URL.
   * @param $subscriber
   *   The subscriber URL that is the callback to be invoked when the topic
   *   changes.
   * @param $secret
   *   Secret for message authentication.
   */
  public function save($topic, $subscriber, $secret);

  /**
   * Delete a subscription.
   *
   * @param $topic
   *   A topic URL.
   * @param $subscriber
   *   The subscriber URL that is the callback to be invoked when the topic
   *   changes.
   */
  public function delete($topic, $subscriber);

  /**
   * Find all subscriber URLs for a given topic URL.
   *
   * @param $topic
   *   A topic URL.
   *
   * @return
   *   An array of subscriber URLs.
   */
  public function all($topic);

  /**
   * Expire subscriptions older than $time.
   *
   * @param $time
   *   Subscriptions older than time() - $time will be deleted.
   */
  public function expire($time);
}
